MODEL_CONTROL_MASKING_LARGE = "control_masking_large"
MODEL_MASKING_LARGE = "masking_large"
MODEL_MASKING_SMALL = "masking_small"

MODELS = [MODEL_CONTROL_MASKING_LARGE, MODEL_MASKING_LARGE, MODEL_MASKING_SMALL]
